class = require"middleclass"

local WorldMap = class("WorldMap")

-- Constants are dealt with as class variables.
-- This way they can be changed for experiments with different world kinds.

WorldMap.static.LatitudeToKM = 111
-- The first day of the year is the spring equinox throughout this file
WorldMap.static.SolarDays = 365
WorldMap.static.HoursInADay = 24
WorldMap.static.AxialTilt = 23.4
WorldMap.static.MaxLatitude = 90
WorldMap.static.TropicalNormalizationTemperature = -10
WorldMap.static.CelsiusPerLatitude = WorldMap.LatitudeToKM/145
WorldMap.static.ElevationMetersPerCelsius = 220
-- The higher the irradiation, the hotter the planet.
WorldMap.MaxIrradiation = 10
-- The below makes it harder for air to exchange temperature with the ground, effectively making
-- the climate milder and reducing seasonal changes. Values below 4 tend to generate artifacts.
WorldMap.HeatCapacityCoefficient = 7
-- The below is how quickly the atmosphere requlates the air temperature back to base temperatures
WorldMap.static.AtmosphereNormalizationFactor = 0.25

function WorldMap:initialize(pX, pY, pTileSize, pStartingLatitude, pPrepareMap)
	self._Definitions = {
		X = pX,
		Y = pY,
		TileSize = pTileSize,
		Latitude = pStartingLatitude, -- southern hemisphere = latitudes below 0
		MinHeight = 0,
		MaxHeight = 0,
		Changed = false,
		-- Day 1 is the day of the spring equinox
		CurrentDay = 1
	}
	self._HeightMap = {}
	for x = 1, pX do
		self._HeightMap[x] = {}
		for y = 1, pY do
			self._HeightMap[x][y] = 0
		end
	end
	if pPrepareMap then
		self:fractalGen()
		self:adjustHeights()
		self:initializeTemperature()
	end
end

function WorldMap:getDimensions()
	return self._Definitions.X, self._Definitions.Y
end

function WorldMap:getX()
	return self._Definitions.X
end

function WorldMap:getY()
	return self._Definitions.Y
end

function WorldMap:getHeight(pX, pY)
	return self._HeightMap[pX][pY]
end

function WorldMap:getTileSize()
	return self._Definitions.TileSize
end

function WorldMap:getLatitude(pY)
	if pY then
		return self._Definitions.Latitude - (pY - 1) * self:getTileSize() / WorldMap.LatitudeToKM
	else
		return self._Definitions.Latitude
	end
end

-- helper function for getting min/max height
local function lRecalculateMinMax(map)
	local lX, lY = map:getDimensions()
	local lMinHeight, lMaxHeight
	for x = 1, lX do
		for y = 1, lY do
			if (x==1) and (y==1) then
				lMinHeight = map:getHeight(x, y)
				lMaxHeight = map:getHeight(x, y)
			elseif map:getHeight(x, y) < lMinHeight then
				lMinHeight = map:getHeight(x, y)
			elseif map:getHeight(x, y) > lMaxHeight then
				lMaxHeight = map:getHeight(x, y)
			end
		end
	end
	map._Definitions.MinHeight = lMinHeight
	map._Definitions.MaxHeight = lMaxHeight
end

function WorldMap:getMinHeight()
	if self._Definitions.Changed then
		self._Definitions.Changed = false
		lRecalculateMinMax(self)
	end
	return self._Definitions.MinHeight
end

function WorldMap:getMaxHeight()
	if self._Definitions.Changed then
		self._Definitions.Changed = false
		lRecalculateMinMax(self)
	end
	return self._Definitions.MaxHeight
end

function WorldMap:setHeight(pX, pY, pHeight)
	self._HeightMap[pX][pY] = pHeight
	self._Definitions.Changed = true
end

function WorldMap:getDay()
	return self._Definitions.CurrentDay
end

function WorldMap:addDays(pDays)
	self._Definitions.CurrentDay = self._Definitions.CurrentDay + pDays
	if self._Definitions.CurrentDay > WorldMap.SolarDays then
		self._Definitions.CurrentDay = self._Definitions.CurrentDay - WorldMap.SolarDays
	end
end



-- FRACTAL GENERATION --

-- helper function selecting two random points to define the line
local function lSetLine(pX, pY)
	local lLinePoint1 = {math.random(1, pX); math.random(1, pY)}
	local lLinePoint2
	-- make sure the two points are not equal
	repeat 
		lLinePoint2 = {math.random(1, pX); math.random(1, pY)}
	until (lLinePoint1[1] ~= lLinePoint2[1]) or (lLinePoint1[2] ~= lLinePoint2[2])
	-- return the points
	return lLinePoint1, lLinePoint2
end

-- helper function checking on which side of the line a given point is
local function lCheckSide(pLinePoint1, pLinePoint2, pX, pY)
	local lCrossProduct = (pLinePoint2[1]-pLinePoint1[1]) * (pY-pLinePoint1[2]) - (pLinePoint2[2]-pLinePoint1[2]) * (pX-pLinePoint1[1])
	if lCrossProduct == 0 then
		return "on the line"
	elseif lCrossProduct > 0 then
		return "right"
	else
		return "left"
	end
end


function WorldMap:fractalGen(pIterations, pMaxChange, pSeed)
	-- Generate a height map with the following algorithm:
	-- draw a line across the map, elevate terrain on one side of it
	-- repeat a lot of times...
	-- The drawback of this method is that it doesn't generate ocean around the map.
	math.randomseed(pSeed or os.time())
	pIterations = pIterations or 10000
	pMaxChange = pMaxChange or 10
	
	for i = 1, pIterations do
		local lLinePoint1, lLinePoint2 = lSetLine(self:getDimensions())
		local lChange = math.random(1, pMaxChange)
		for x = 1, self:getX() do
			for y = 1, self:getY() do
				if lCheckSide(lLinePoint1, lLinePoint2, x, y) == "right" then
					self:setHeight(x, y, self:getHeight(x, y) + lChange)
				end
			end
		end
	end
	self._Definitions.MinHeight = lMinHeight
	self._Definitions.MaxHeight = lMaxHeight
end

-- END of FRACTAL GENERATION --

-- ADJUSTING HEIGHTS --

function WorldMap:adjustHeights(pSeaLevel, pMaxHeight)
	pSeaLevel = pSeaLevel or 0.5
	pMaxHeight = pMaxHeight or 5000
	local lMinHeight = self:getMinHeight()
	local lMaxHeight = self:getMaxHeight()
	local lSpread = lMaxHeight - lMinHeight
	local lSea = lMinHeight + pSeaLevel * lSpread
	local lChange = pMaxHeight / (lMaxHeight - lSea)
	for x = 1, self:getX() do
		for y = 1, self:getY() do
			local lHeight = self:getHeight(x, y)
			self:setHeight(x, y, (lHeight - lSea) * lChange)
		end
	end
end

-- END of ADJUSTING HEIGHTS --

-- load map parameters from a file or from a string
function WorldMap:ingest(pMapString)
end

-- CREATE TEMPERATURE DATA --

--TO DO add changes based on current temperature
function WorldMap:getHeatCapacity(pX, pY)
	local lHeatCapacity, lDMulti, lDepthCapacity
	local WaterCoefficient = 5
	local LandCoefficient = 1
	if self:isOcean(pX, pY) then
		-- 4 C is the temperature at which water is the densest
		-- deeper waters have a stronger cooling effect
		lDepthCapacity = math.abs(self:getHeight(pX, pY)) / 1000
		lDMulti = math.abs(4 - self:getGroundTemperature(pX, pY)) / 10
		lHeatCapacity = WaterCoefficient * WorldMap.HeatCapacityCoefficient 
	else
		lDepthCapacity = 0
		lDMulti = math.abs(self:getBaseTemperature(pX, pY) - self:getGroundTemperature(pX, pY)) / 20
		lHeatCapacity = LandCoefficient * WorldMap.HeatCapacityCoefficient
	end
	return lHeatCapacity * (1 + lDMulti + lDepthCapacity)
end

function WorldMap:getAirTemperature(pX, pY)
	return self._AirTemperature[pX][pY]
end

function WorldMap:getGroundTemperature(pX, pY)
	return self._GroundTemperature[pX][pY]
end

function WorldMap:getBaseTemperature(pX, pY)
	return self._BaseTemperature[pX][pY]
end

function WorldMap:isOcean(pX, pY)
	if self:getHeight(pX, pY) < 0 then
		return true
	else
		return false
	end
end

function WorldMap:getSunAngleAtNoon(pY, pDay)
	-- get a "season" from -1 for the winter solstice to + 1 for the summer solstice
	-- operate on the winter solstice:
	pDay = pDay + WorldMap.SolarDays/4
	if pDay > WorldMap.SolarDays then
		pDay = pDay - WorldMap.SolarDays
	end
	local lSeason 
	if pDay <= WorldMap.SolarDays/2 then
		lSeason = pDay / WorldMap.SolarDays * 4 - 1
	else
		lSeason = (WorldMap.SolarDays - pDay) / WorldMap.SolarDays * 4 - 1
	end
	local lLat = self:getLatitude(pY)
	if lLat < 0 then
		lLat = -lLat
		lSeason = -lSeason
	end
	local lAngle = WorldMap.MaxLatitude - lLat + WorldMap.AxialTilt * lSeason
	if lAngle > WorldMap.MaxLatitude then
		lAngle = 2 * WorldMap.MaxLatitude - lAngle
	end
	return lAngle
end

function WorldMap:getSolstices(pLatitude)
	local lSummer = WorldMap.SolarDays / 4
	local lWinter = WorldMap.SolarDays * 3 / 4
	if pLatitude < 0 then
		lSummer, lWinter = lWinter, lSummer	
	end
	return lSummer, lWinter
end

function WorldMap:getDayLength(pY, pDay)
	pDay = pDay or self:getDay()
	local lLatitude = self:getLatitude(pY)
	local lSolsticeToSolstice = WorldMap.SolarDays / 2
	local lEquinoxToSolstice = lSolsticeToSolstice / 2
	local lEquinoxDayLength = WorldMap.HoursInADay / 2
	local lDayDelta = WorldMap.HoursInADay / 2 * lLatitude / 90 / lSolsticeToSolstice
	local lDaysFromSolstice = (pDay + lEquinoxToSolstice) % lSolsticeToSolstice

	local lSolsticeDelta, lChange
	if (pDay > lEquinoxToSolstice) and (pDay < lEquinoxToSolstice * 3) then
		lSolsticeDelta = lEquinoxToSolstice * lDayDelta
		lChange = -1
	else
		lSolsticeDelta = - lEquinoxToSolstice * lDayDelta
		lChange = 1
	end
	if lLatitude < 0 then
		lSolsticeDelta = -lSolsticeDelta
		lChange = -lChange
	end
	return lEquinoxDayLength + lSolsticeDelta + lDaysFromSolstice * lChange * lDayDelta
end

function WorldMap:calculateIrradiation(pY, pDay)
	pDay = pDay or self:getDay()
	local lAngle = self:getSunAngleAtNoon(pY, pDay)
	local lDayCoefficient = self:getDayLength(pY, pDay) / WorldMap.HoursInADay
	return (math.max(0, lAngle / WorldMap.MaxLatitude) * WorldMap.MaxIrradiation * lDayCoefficient)^2
end

function WorldMap:heatTile(pX, pY, pIrradiation)
	local lCapacity = self:getHeatCapacity(pX, pY)
	-- Elevate temperature by irradiation
	self._GroundTemperature[pX][pY] = self:getGroundTemperature(pX, pY) + pIrradiation / lCapacity
	-- Exchange temperature between land and air
	local lDTemperature = self:getGroundTemperature(pX, pY) - self:getAirTemperature(pX, pY)
	self._GroundTemperature[pX][pY] = self:getGroundTemperature(pX, pY) - lDTemperature / lCapacity
	self._AirTemperature[pX][pY] = self._AirTemperature[pX][pY] + lDTemperature 
	-- Normalizing/cooling - this is the air going to the upper partns of the atmosphere --
	local lDNormal = self:getBaseTemperature(pX, pY) - self:getAirTemperature(pX, pY)
	self._AirTemperature[pX][pY] = self:getAirTemperature(pX, pY) + lDNormal * WorldMap.AtmosphereNormalizationFactor
end

function WorldMap:runHeatMap(pPeriod)
	for n=1, pPeriod do
		self:addDays(1)
		for y=1, self:getY() do
			local lIrradiation = self:calculateIrradiation(y)
			for x=1, self:getX() do
				self:heatTile(x, y, lIrradiation, pPeriod)
			end
		end
		local lAir = {}
		for x=1, self:getX() do
			lAir[x] = {}
			for y=1, self:getY() do
				lAir[x][y] = self._AirTemperature[x][y]
			end
		end
	
		for x=1, self:getX() do
			for y=1, self:getY() do
				local LDTemperature
				if x<self:getX() then
					lDTemperature = self._AirTemperature[x][y] - self._AirTemperature[x+1][y]
					lAir[x+1][y] = lAir[x+1][y] + lDTemperature
					-- ADD WIND FOR AIR MOVEMENT HERE --
				else
					lDTemperature = 0
				end
				lAir[x][y] = lAir[x][y] - lDTemperature
				-- ADD WIND FOR AIR MOVEMENT HERE --
				if y<self:getY() then
					lDTemperature = self._AirTemperature[x][y] - self._AirTemperature[x][y+1]
					lAir[x][y+1] = lAir[x][y+1] + lDTemperature
					-- ADD WIND FOR AIR MOVEMENT HERE --
				else
					lDTemperature = 0
				end
				lAir[x][y] = lAir[x][y] - lDTemperature
				-- ADD WIND FOR AIR MOVEMENT HERE --
			end
		end
		self._AirTemperature = lAir
	end	
end

function WorldMap:initializeTemperature()
	-- create base temperatures for each latitude
	self._LatitudeTemperature = {}
	for y=1, self:getY() do
		self._LatitudeTemperature[y] = WorldMap.TropicalNormalizationTemperature
--		if math.abs(self:getLatitude(y)) < WorldMap.AxialTilt then
--			self._LatitudeTemperature[y] = WorldMap.TropicalNormalizationTemperature
--		else
--			self._LatitudeTemperature[y] = WorldMap.TropicalNormalizationTemperature 
--				- (math.abs(self:getLatitude(y)) - WorldMap.AxialTilt) * WorldMap.CelsiusPerLatitude
--		end
	end
	-- create base temperatures for every cell based on elevation and latitude
	self._GroundTemperature ={}
	self._AirTemperature = {}
	self._BaseTemperature = {}
	for x=1, self:getX() do
		self._GroundTemperature[x] = {}
		self._AirTemperature[x] = {}
		self._BaseTemperature[x] = {}
		for y=1, self:getY() do
			local height = math.max(0, self:getHeight(x, y))
			self._BaseTemperature[x][y] = self._LatitudeTemperature[y]
				- height / WorldMap.ElevationMetersPerCelsius
			self._AirTemperature[x][y] = self._BaseTemperature[x][y]
			self._GroundTemperature[x][y] = self._BaseTemperature[x][y]
		end
	end
end


-- END of TEMPERATURE DATA --


--
--local function checkValues(pValue, pAllowedValues)
--	local lOk = false
--	for k, v in pairs(pAllowedValues) do
--		if pValue == v then
--			lOk = true
--		end
--	end
--	if not lOk then
--		error(toString(pValue) .. " is not a valid value.", 3)
--	end
--end

return WorldMap

